/*
    c99tree - parse tree utility
    Copyright (C) 2010  Tibor Palinkas

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "tree.h"
#include "tree_print.h"
#include "c99tree.h"
#include "genht/htsi.h"
#include "genht/hash.h"

const char *node_type_names[] = {
"CALL", "STATEMENT", "TYPE", "DECL", "FUNCDEF", "ARGLIST", "ARG",
"BLOCK", "CLASS", "TYPEOF", "ATTRLIST", "ATTR", "PTR", "INDEX",
"QUALIFIER", "ID", "ELLIPSIS", "ENUM", "INIT", "STRUCT", "BITS", "ASM",
"LABEL", "CONDITION", "THEN", "ELSE", "OP", "ICONST",
"FCONST", "SCONST", "SCONSTLIST", "DEBUG", "UNION", "PRAGMA", "TYPECUSTOM",
"TYPESPEC", "TYPENAME", "TYPEDEF",  "EXPRLIST", "COMMENT", "INITLIST", NULL};

htsi_t *node_type_hash = NULL;

node_type_t Ntype(const char *typenam)
{
	const htsi_entry_t *ent;
	ent = (const htsi_entry_t *)htsi_getentry(node_type_hash, (htsi_key_t)typenam);
	if (ent == NULL)
		return T_INVALID;
	return (node_type_t)ent->value;
}

/* --- hash --- */
static int keyeq(char *a, char *b)
{
	return !strcmp(a, b);
}

static unsigned int hash(char *key)
{
	return murmurhash(key, strlen(key));
}


void tree_init()
{
	int n;

	node_type_hash = htsi_alloc(hash, keyeq);

	for(n = 0; n < T_INVALID; n++)
		htsi_set(node_type_hash, (htsi_key_t)node_type_names[n], n);
}

void tree_uninit()
{
	htsi_free(node_type_hash);
}


node_t *Nalloc(node_type_t type, char *name, node_t *child)
{
	node_t *n;

	n = malloc(sizeof(node_t));
	n->type = type;
	n->name = name;
	n->next = NULL;
	n->children = child;
	loc_cpy(&(n->loc), loc_current);
	return n;
}

node_t *Nappend(node_t *dest, node_t *src)
{
	node_t *odest;

	odest = dest;

	for(; dest->next != NULL; dest = dest->next) ;
	dest->next = src;

	return odest;
}

static void Nprint_(const char *prefix, node_t *node)
{
	char *pr2;
	node_t *n;
	int l, r;

	r = Nprint_cb(prefix, node, PS_OPENING);
	if ((r == 0) && (node->children != NULL)) {
		l = strlen(prefix);
		pr2 = malloc(l + 2);
		*pr2 = *prefix;
		memcpy(pr2+1, prefix, l+1);
		for(n = node->children; n != NULL; n = n->next)
			Nprint_(pr2, n);
		free(pr2);
	}
	Nprint_cb(prefix, node, PS_CLOSING);
}

void Nprint(const char *prefix, node_t *node)
{
	node_t *n;
	for(n = node; n != NULL; n = n->next)
		Nprint_(prefix, n);
}

node_t *Nop(const char *op, node_t *o1, node_t *o2)
{
	node_t *res;

	res = Nalloc(T_OP, strdup(op), o1);
	if (o2 != NULL)
		Nappend(o1, o2);
	return res;
}

node_t *Nclone(const node_t *node)
{
	node_t **prev, *t, *ret;
	const node_t *n;

	if (node == NULL)
		return NULL;

	prev = &ret;
	for(n = node; n != NULL; n = n->next) {
		*prev = t = Nalloc(n->type, n->name == NULL ? NULL : strdup(n->name), Nclone(n->children));
		prev = &(t->next);
	}
	*prev = NULL;
	return ret;
}

void Nfree(node_t *node)
{
	node_t *next;
	node_t *n;

	for(n = node; n != NULL; n = next) {
		if (n->children != NULL)
			Nfree(n->children);
		next = n->next;
		if (n->name != NULL)
			free(n->name);
		free(n);
	}
}
