typedef struct node_s node_t;

#include "location.h"

/* tree node types */
typedef enum node_type_e {
	T_CALL,          /* function call instruction */
	T_STATEMENT,
	T_TYPE,          /* type container for qualifiers, specifier, typenames, attribs */
	T_DECL,
	T_FUNCDEF,
	T_ARGLIST,       /* function arguments */
	T_ARG,           /* a specific argument */
	T_BLOCK,         /* compound statement */
	T_CLASS,
	T_TYPEOF,
	T_ATTRLIST,
	T_ATTR,
	T_PTR,
	T_INDEX,
	T_QUALIFIER,
	T_ID,
	T_ELLIPSIS,      /* ... */
	T_ENUM,
	T_INIT,
	T_STRUCT,
	T_BITS,          /* :X part of a bitfield */
	T_ASM,
	T_LABEL,         /* jump and switch labels */
	T_CONDITION,     /* condition section of while(), for() and if() */
	T_THEN,
	T_ELSE,
	T_OP,            /* operator */
	T_ICONST,        /* integer constant */
	T_FCONST,        /* floating point constant */
	T_SCONST,        /* string constant */
	T_SCONSTLIST,    /* appended string constants */
	T_DEBUG,         /* dummy node for debugging - shouldn't show up in any of the production output */
	T_UNION,
	T_PRAGMA,
	T_TYPECUSTOM,    /* typedef'd types */
	T_TYPESPEC,      /* long, short, signed, unsigned, _Complex and _Imaginary */
	T_TYPENAME,      /* name of the type (like int, float, etc.) */
	T_TYPEDEF,
	T_EXPRLIST,      /* comma separated list of expressions (1,2,3) */
	T_COMMENT,
	T_INITLIST,      /* initializer list (int a[] = {1, 2}; ) */

	T_INVALID        /* matches with the NULL in the name list */
} node_type_t;

extern const char *node_type_names[];

struct node_s {
	node_type_t type;

	char *name;         /* name field - textual payload for some nodes */
	loc_t loc;          /* location info */
	int print_data;     /* random data for the output layer */

	node_t *next;
	node_t *children;
};

/* builds/frees type hash */
void tree_init();
void tree_uninit();


/* Look up type by name */
node_type_t Ntype(const char *typenam);

/* Allocate a new tree node - ->next is empty, line is filled in from global lineno */
node_t *Nalloc(node_type_t type, char *name, node_t *child);

/* Append src node to the last ->next starting from dest - returns dest */
node_t *Nappend(node_t *dest, node_t *src);

/* Create operator node with 2 operands; 2nd operand may be NULL */
node_t *Nop(const char *op, node_t *o1, node_t *o2);

/* Clone a subtree recursively */
node_t *Nclone(const node_t *node);

/* Free a subtree recurisively */
void Nfree(node_t *node);

/* Print node and all ->next nodes recursively, prefix lines with prefix */
void Nprint(const char *prefix, node_t *node);

/* Append src node as dest node's last child
   Make copy of dest_ and src_ as they may be calls */
#define Nchappend(dest_, src_)	\
	do { \
		node_t *dest, *src; \
		dest = dest_; \
		src = src_; \
		if ((dest)->children != NULL) \
			Nappend((dest)->children, (src)); \
		else \
			(dest)->children = src; \
	} while(0)


/* Insert src node as dest node's first child
   Make copy of dest_ and src_ as they may be calls */
#define Nchinsert(dest_, src_)	\
	do { \
		node_t *dest, *src; \
		dest = dest_; \
		src = src_; \
		if ((dest)->children != NULL) {\
			if (src != NULL) { \
				Nappend(src, (dest)->children); \
				(dest)->children = src; \
			} \
		} \
		else \
			(dest)->children = src; \
	} while(0)


#define NTYinsert(dest_, src_) \
	do { \
		node_t *dest, *src; \
		dest = dest_; \
		src = src_; \
		if ((dest->children == NULL) || (dest->children->type != T_TYPE)) \
			dest->children = Nalloc(T_TYPE, NULL, dest->children); \
		src->next = dest->children->children; \
		dest->children->children = src; \
	} while(0)
